@@ -29,9 +29,14 @@
  */
 
 /*
- * $Id: ipsec_tunnel.c,v 0.4 1997/01/15 01:28:15 ji Rel $
+ * $Id: ipsec_tunnel.c,v 0.5 1997/06/03 04:24:48 ji Rel $
  *
  * $Log: ipsec_tunnel.c,v $
+ * Revision 0.5  1997/06/03 04:24:48  ji
+ * Added transport mode.
+ * Changed the way routing is done.
+ * Lots of bug fixes.
+ *
  * Revision 0.4  1997/01/15 01:28:15  ji
  * No changes.
  *
@@ -137,27 +142,123 @@
 
 
 #ifdef DEBUG_IPSEC_TUNNEL
-int debug_tunnel = 0;
+int debug_tunnel = -1;
 #endif
 
-/* 
- *	Our header is a simple IP packet with no options
- */
  
-#define tunnel_hlen	sizeof(struct iphdr)
+#define TUNL_HLEN	96		/* should be 96 XXXXX big enough, multiple of 8 */
 
 /*
- *	Okay, this needs to be high enough that we can fit a "standard"
- *	ethernet header and an IP tunnel header into the outgoing packet.
- *	[36 bytes]
- *
+ *	This is mostly skbuff.c:skb_copy().
  */
  
-#define TUNL_HLEN	(((ETH_HLEN+15)&~15))
+struct sk_buff *skb_copy_expand(struct sk_buff *skb, int priority, int headroom, int tailroom)
+{
+	struct sk_buff *n;
+	unsigned long offset;
 
+	/*
+	 *	Allocate the copy buffer
+	 */
+	 
+	IS_SKB(skb);
+	
+	n=alloc_skb(skb->truesize-sizeof(struct sk_buff) + headroom+tailroom, priority);
+	if(n==NULL)
+		return NULL;
+
+	/*
+	 *	Shift between the two data areas in bytes
+	 */
+	 
+	/* offset=n->head-skb->head; */ /* moved down a few lines */
+
+	/* Set the data pointer */
+	skb_reserve(n,skb->data-skb->head+headroom);
+	/* Set the tail pointer and length */
+	skb_put(n,skb->len);
+
+	offset=n->head + headroom - skb->head;
+
+	/* Copy the bytes */
+	memcpy(n->head + headroom, skb->head,skb->end-skb->head);
+	n->link3=NULL;
+	n->list=NULL;
+	n->sk=NULL;
+	n->when=skb->when;
+	n->dev=skb->dev;
+	n->h.raw=skb->h.raw+offset;
+	n->mac.raw=skb->mac.raw+offset;
+	n->ip_hdr=(struct iphdr *)(((char *)skb->ip_hdr)+offset);
+	n->saddr=skb->saddr;
+	n->daddr=skb->daddr;
+	n->raddr=skb->raddr;
+	n->seq=skb->seq;
+	n->end_seq=skb->end_seq;
+	n->ack_seq=skb->ack_seq;
+	n->acked=skb->acked;
+	memcpy(n->proto_priv, skb->proto_priv, sizeof(skb->proto_priv));
+	n->used=skb->used;
+	n->free=1;
+	n->arp=skb->arp;
+	n->tries=0;
+	n->lock=0;
+	n->users=0;
+	n->pkt_type=skb->pkt_type;
+	n->stamp=skb->stamp;
+	
+	IS_SKB(n);
+	return n;
+}
+
+#ifdef REAL_LOCKING_P
+/*
+ *	Locking
+ */
+ 
+static int ipsec_tunnel_lock(struct ipsecpriv *prv)
+{
+	unsigned long flags;
+	save_flags(flags);
+	cli();
+	/*
+	 *	Lock in an interrupt may fail
+	 */
+	if(prv->locked && in_interrupt())
+	{
+		restore_flags(flags);
+		return 0;
+	}
+	while(prv->locked)
+		sleep_on(&prv->wait_queue);
+	prv->locked=1;
+	restore_flags(flags);
+	return 1;
+}
+
+static void ipsec_tunnel_unlock(struct ipsecpriv *prv)
+{
+	prv->locked=0;
+	wake_up(&prv->wait_queue);
+}
+#endif
 
 static int ipsec_tunnel_open(struct device *dev)
 {
+	struct ipsecpriv *prv = dev->priv;
+	
+	/*
+	 * Can't open until attached.
+	 */
+
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_INIT)
+		printk("ipsec_tunnel_open: prv->dev = %s\n", prv->dev?prv->dev->name:"NONE");
+#endif
+
+	if (prv->dev == NULL)
+		return -ENODEV;
+	
 	MOD_INC_USE_COUNT;
 	return 0;
 }
@@ -168,6 +269,224 @@ static int ipsec_tunnel_close(struct device *dev)
 	return 0;
 }
 
+/*
+ * Revectored calls.
+ * For each of these calls, a field exists in our private structure.
+ */
+
+static int ipsec_tunnel_start_xmit(struct sk_buff *skb, struct device *dev)
+{
+	struct ipsecpriv *prv = dev->priv;
+	int ret;
+	
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_REVEC)
+		printk("ipsec_tunnel: Revectored start_xmit\n");
+#endif
+
+	cli();
+	if (dev->tbusy != 0) 
+	{
+		sti();
+		/* stats->tx_errors++; */
+		printk("ipsec_tunnel_start_xmit: busy, returning\n");
+		return 1;
+	}
+	dev->tbusy = 1;
+	sti();
+
+	if (dev->hard_header_len == 14 &&
+	    skb->data[12] == 0x08 &&
+	    skb->data[13] == 0x00)
+	{
+#ifdef DEBUG_IPSEC_TUNNEL
+		if (debug_tunnel & DB_TN_REVEC)
+			printk("ipsec_tunnel_start_xmit: we have an IP packet on an ethernet\n");
+#endif
+		ret = ipsec_tunnel_do_xmit(skb, dev);
+	}
+	else
+	{
+		struct sk_buff *newskb=skb_clone(skb, GFP_ATOMIC);
+		if(newskb)
+		{
+			newskb->dev=prv->dev;
+			newskb->arp=1;
+			ret = prv->hard_start_xmit(newskb, prv->dev);
+		}
+		else
+		{
+			ret = -1;
+		}
+	}
+	dev_kfree_skb(skb, FREE_WRITE);
+	dev->tbusy = 0;
+	return ret;
+}
+
+
+
+static struct enet_statistics *ipsec_tunnel_get_stats(struct device *dev)
+{
+	return &(((struct ipsecpriv *)(dev->priv))->mystats);
+}
+
+static int ipsec_tunnel_hard_header(struct sk_buff *skb, struct device *dev,
+	unsigned short type, void *daddr, void *saddr, unsigned len)
+{
+	struct ipsecpriv *prv = dev->priv;
+	
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_REVEC)
+		printk("ipsec_tunnel: Revectored hard_header\n");
+#endif
+	return prv->hard_header(skb, prv->dev, type, daddr, saddr, len);
+}
+
+static int ipsec_tunnel_rebuild_header(void *eth, struct device *dev,
+			unsigned long raddr, struct sk_buff *skb)
+
+{
+	struct ipsecpriv *prv = skb->dev->priv;
+	struct device *tmp;
+	int ret;
+	
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_REVEC)
+		printk("ipsec_tunnel: Revectored rebuild_header\n");
+#endif
+	tmp = skb->dev;
+	skb->dev = prv->dev;
+	
+	ret = prv->rebuild_header(eth, prv->dev, raddr, skb);
+	skb->dev = tmp;
+	return ret;
+}
+
+static int ipsec_tunnel_set_mac_address(struct device *dev, void *addr)
+{
+	struct ipsecpriv *prv = dev->priv;
+	
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_REVEC)
+		printk("ipsec_tunnel: Revectored set_mac_address_\n");
+#endif
+	return prv->set_mac_address(prv->dev, addr);
+
+}
+
+static void ipsec_tunnel_cache_bind(struct hh_cache **hhp, struct device *dev,
+				 unsigned short htype, __u32 daddr)
+{
+	struct ipsecpriv *prv = dev->priv;
+	
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_REVEC)
+		printk("ipsec_tunnel: Revectored cache_bind\n");
+#endif
+	prv->header_cache_bind(hhp, dev, htype, daddr);
+	return;
+}
+
+static void ipsec_tunnel_cache_update(struct hh_cache *hh, struct device *dev, unsigned char *  haddr)
+{
+	struct ipsecpriv *prv = dev->priv;
+	
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_REVEC)
+		printk("ipsec_tunnel: Revectored cache_update\n");
+#endif
+	prv->header_cache_update(hh, prv->dev, haddr);
+	return;
+}
+
+/*
+ * We call the attach routine to attach another device.
+ */
+
+static int ipsec_tunnel_attach(struct device *tndev, struct ipsecpriv *prv, struct device *dev)
+{
+	prv->dev = dev;
+	prv->hard_start_xmit = dev->hard_start_xmit;
+	prv->get_stats = dev->get_stats;
+
+	if (dev->hard_header)
+	{
+		prv->hard_header = dev->hard_header;
+		tndev->hard_header = ipsec_tunnel_hard_header;
+	}
+	else
+		tndev->hard_header = NULL;
+	
+	if (dev->rebuild_header)
+	{
+		prv->rebuild_header = dev->rebuild_header;
+		tndev->rebuild_header = ipsec_tunnel_rebuild_header;
+	}
+	else
+		tndev->rebuild_header = NULL;
+	
+	if (dev->set_mac_address)
+	{
+		prv->set_mac_address = dev->set_mac_address;
+		tndev->set_mac_address = ipsec_tunnel_set_mac_address;
+	}
+	else
+		tndev->set_mac_address = NULL;
+	
+	if (dev->header_cache_bind)
+	{
+		prv->header_cache_bind = dev->header_cache_bind;
+		tndev->header_cache_bind = ipsec_tunnel_cache_bind;
+	}
+	else
+		tndev->header_cache_bind = NULL;
+
+	if (dev->header_cache_update)
+	{
+		prv->header_cache_update = dev->header_cache_update;
+		tndev->header_cache_update = ipsec_tunnel_cache_update;
+	}
+	else
+		tndev->header_cache_update = NULL;
+
+	tndev->hard_header_len = dev->hard_header_len /* XXX + TUNL_HLEN */;
+	tndev->mtu = dev->mtu - TUNL_HLEN;
+	/* tndev->type = dev->type; should remain tunnel */
+	tndev->addr_len = dev->addr_len;
+	
+	return 0;
+}
+
+static int ipsec_tunnel_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
+{
+	struct ipsectunnelconf *cf = (struct ipsectunnelconf *)&ifr->ifr_data;
+	struct ipsecpriv *prv = dev->priv;
+	struct device *them = dev_get(cf->cf_name);
+	
+	switch (cmd)
+	{
+	case IPSEC_SET_DEV:
+		if (them == NULL)
+		{
+#ifdef DEBUG_IPSEC_TUNNEL
+			if (debug_tunnel & DB_TN_INIT)
+				printk("ipsec_tunnel_ioctl: them is null\n");
+#endif
+
+			return -ENODEV;
+		}
+		
+		if (prv->dev)
+			return -EBUSY;
+		return ipsec_tunnel_attach(dev, dev->priv, them);
+
+	default:
+		return -EOPNOTSUPP;
+	}
+}
+
+
 #ifdef DEBUG_IPSEC_TUNNEL
 static void print_ip(struct iphdr *ip)
 {
@@ -193,33 +512,36 @@ static void print_ip(struct iphdr *ip)
  */
 
 static int
-ipsec_tunnel_xmit(struct sk_buff *skb, struct device *dev)
+ipsec_tunnel_do_xmit(struct sk_buff *oskb, struct device *odev)
 {
+	struct device *dev;			/* Our device */
+	struct ipsecpriv *prv;
+	struct sk_buff *skb;
 	struct enet_statistics *stats;		/* This device's statistics */
 	struct rtable *rt;     			/* Route to the other host */
 	struct device *tdev;			/* Device to other host */
 	struct iphdr  *iph;			/* Our new IP header */
 	__u32          target;			/* The other host's IP address */
+	__u32	orgdst;			/* Original IP destinatin address */
 	int	iphlen;
 	int	pyldsz;
 	int      max_headroom;			/* The extra header space needed */
 	int	max_tailroom;		/*  The extra stuffing needed */
-
+	__u8	savedheader[16];	/* should be plenty of room */
+	int i;
+	int ret;
 	
+
 	struct sockaddr_encap matcher;
 	struct eroute *er;
 	struct tdb *tdbp, *tdbq;
 	int oerror;
 	
-#ifdef DEBUG_IPSEC_TUNNEL
-#ifdef DEBUG_PP
-	outb(0x01, 0x378);
-#endif
-#endif
+
 	/*
 	 *	Return if there is nothing to do.  (Does this ever happen?)
 	 */
-	if (skb == NULL || dev == NULL) {
+	if (oskb == NULL || odev == NULL) {
 #ifdef DEBUG_IPSEC_TUNNEL
 		if (debug_tunnel & DB_TN_XMIT)
 			printk ( KERN_INFO "tunnel: Nothing to do!\n" );
@@ -227,24 +549,65 @@ ipsec_tunnel_xmit(struct sk_buff *skb, struct device *dev)
 		return 0;
 	}
 
-	/* 
-	 *	Make sure we are not busy (check lock variable) 
-	 */
-	 
-	stats = (struct enet_statistics *)dev->priv;
+	prv = odev->priv;
+	dev = prv->dev;
+
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_XMIT)
+	{
+		int i;
+		char c;
+		
+		printk("do_xmit: hard_header_len = %d", dev->hard_header_len);
+		c = ' ';
+		
+		for (i=0; i < dev->hard_header_len; i++)
+		{
+			printk("%c%02x", c, oskb->data[i]);
+			c = ':';
+		}
+		printk("\n");
+	}
+#endif
 	cli();
 	if (dev->tbusy != 0) 
 	{
 		sti();
-		stats->tx_errors++;
-		return(1);
+		/* stats->tx_errors++; */
+		printk("ipsec_tunnel_do_xmit: busy, returning\n");
+		return 1;
 	}
 	dev->tbusy = 1;
 	sti();
   
-	iph = (struct iphdr *) skb->data;
+
+	skb = skb_copy_expand(oskb, GFP_ATOMIC, 64, 32);
+	if (skb)
+	{
+		skb->dev = prv->dev;
+		skb->arp = 1;
+	}
+	else
+		return 0;
+
+	iph = (struct iphdr *) (skb->data + dev->hard_header_len);
 	iphlen = iph->ihl << 2;
 	pyldsz = ntohs(iph->tot_len) - iphlen;
+
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_XMIT)
+	{
+		printk("ipsec_tunnel_do_xmit: packet contents:");
+		for (i = 0; i < skb->len; i++)
+			printk(" %02x", skb->data[i]);
+		printk("\n");
+		printk("\tskb->len=%d\n", skb->len);
+	}
+#endif
+
+	
+
+
 	
 	/*
 	 * First things first -- look us up in the erouting tables.
@@ -260,19 +623,32 @@ ipsec_tunnel_xmit(struct sk_buff *skb, struct device *dev)
 	er = ipsec_findroute(&matcher);
 	if (er == NULL)
 	{
-		printk("%s: Packet with no e-route!\n", dev->name);
+		printk("%s: Packet [%x->%x] with no e-route!\n", dev->name, ntohl(iph->saddr), ntohl(iph->daddr));
 		dev->tbusy=0;
-		stats->tx_errors++;
-		dev_kfree_skb(skb, FREE_WRITE);
-		return 0;
+		return prv->hard_start_xmit(skb, dev);
 	}
 
+	orgdst = iph->daddr;
+	
+	for (i = 0; i < dev->hard_header_len; i++)
+		savedheader[i] = skb->data[i];
+	skb_pull(skb, dev->hard_header_len);
+
 #ifdef DEBUG_IPSEC_TUNNEL
-#ifdef DEBUG_PP
-	outb(0x03, 0x378);
-#endif
+	if (debug_tunnel & DB_TN_XMIT)
+	{
+		printk("ipsec_tunnel_do_xmit: packet contents after pull:");
+		for (i = 0; i < skb->len; i++)
+			printk(" %02x", skb->data[i]);
+		printk("\n");
+		printk("\tskb->len=%d\n", skb->len);
+	}
 #endif
 
+
+	target = er->er_dst.s_addr;
+
+#ifdef BADCODE
 	/*
 	 *  Now, Look up the tunnel (outer) destination address in the 
 	 *  routing tables
@@ -329,28 +705,15 @@ ipsec_tunnel_xmit(struct sk_buff *skb, struct device *dev)
 		return 0;
 	}
 
-#ifdef DEBUG_IPSEC_TUNNEL
-#ifdef DEBUG_PP
-	outb(0x00, 0x378);
-	outb(0x07, 0x378);
-#endif
-#endif
-
-#ifdef DEBUG_IPSEC_TUNNEL
-	if (debug_tunnel & DB_TN_OHDR)
-	{
-		printk("Old IP Header....\n");
-		print_ip(iph);
-	}
-#endif
+#endif BADCODE
 
 	tdbp = gettdb(er->er_spi, er->er_dst);
 	if (tdbp == NULL)
 	{
 		printk("ipsec_tunnel_xmit: no tdb for spi=%x target=%08x\n", (u_int)ntohl(er->er_spi), (u_int)ntohl(er->er_dst.s_addr));
-		ip_rt_put(rt);
+		/* ip_rt_put(rt); */
 		dev->tbusy=0;
-		stats->tx_errors++;
+		/* stats->tx_errors++; */
 		dev_kfree_skb(skb, FREE_WRITE);
 		return 0;
 	}
@@ -373,7 +736,7 @@ ipsec_tunnel_xmit(struct sk_buff *skb, struct device *dev)
 	/*
 	 * Okay, now see if we can stuff it in the buffer as-is.
 	 */
-	max_headroom += ((tdev->hard_header_len+15)&~15);
+	max_headroom += ((dev->hard_header_len+15)&~15); /* XXX was tdev */
 
 #ifdef DEBUG_IPSEC_TUNNEL
 	if (debug_tunnel & DB_TN_CROUT)
@@ -387,16 +750,17 @@ ipsec_tunnel_xmit(struct sk_buff *skb, struct device *dev)
 	if ((skb_headroom(skb) >= max_headroom) && 
 	    (skb_tailroom(skb) >= max_tailroom) && skb->free)
 	{
-
 #ifdef DEBUG_IPSEC_TUNNEL
 		if (debug_tunnel & DB_TN_CROUT)
 			printk("data fits in existing skb\n");
-#endif		skb->h.iph = (struct iphdr *) skb_push(skb, max_headroom);
-		skb_trim(skb, max_tailroom);
-		skb_device_unlock(skb);
+#else
+		;
+#endif
+
 	}
 	else
 	{
+#ifdef ALLOW_SECOND_SKB_ALLOC
 		struct sk_buff *new_skb;
 
 #ifdef DEBUG_IPSEC_TUNNEL
@@ -436,6 +800,12 @@ ipsec_tunnel_xmit(struct sk_buff *skb, struct device *dev)
 		/* Free the old packet, we no longer need it */
 		dev_kfree_skb(skb, FREE_WRITE);
 		skb = new_skb;
+#else
+		printk("ipsec_tunnel_do_xmit: TRIED TO ALLOCATE %d head and %d tailroom\n", max_headroom, max_tailroom);
+		kfree_skb(skb, FREE_WRITE);
+		dev->tbusy = 0;
+		return 0;
+#endif ALLOW_SECOND_SKB_ALLOC
 	}
 
 	while (tdbp && tdbp->tdb_xform)
@@ -452,34 +822,68 @@ ipsec_tunnel_xmit(struct sk_buff *skb, struct device *dev)
 		tdbp = tdbp->tdb_onext;
 	}
 
-	skb->protocol		=	htons(ETH_P_IP);
+#ifdef DEBUG_IPSEC_TUNNEL
+	if (debug_tunnel & DB_TN_XMIT)
+	{
+		printk("ipsec_tunnel_do_xmit: packet contents after xforms:");
+		for (i = 0; i < skb->len; i++)
+			printk(" %02x", skb->data[i]);
+		printk("\n");
+		printk("\tskb->len=%d\n", skb->len);
+	}
+#endif
 
-	/*
-	 *	Send the packet on its way!
-	 *	Note that dev_queue_xmit() will eventually free the skb.
-	 *	If ip_forward() made a copy, it will return 1 so we can free.
-	 */
 
-	if (ip_forward(skb, dev, IPFWD_NOTTLDEC, target))
-		kfree_skb(skb, FREE_WRITE);
+	if (target != orgdst)
+	{
+		
 
-	/*
-	 *	Clean up:  We're done with the route and the packet
-	 */
+		skb->protocol		=	htons(ETH_P_IP);
+
+		/*
+		 *	Send the packet on its way!
+		 *	Note that dev_queue_xmit() will eventually free the skb.
+		 *	If ip_forward() made a copy, it will return 1 so we can free.
+		 */
+
+#ifdef DEBUG_IPSEC_TUNNEL
+		if (debug_tunnel & DB_TN_INIT)
+		{
+			
+			printk("skb->data = 0x%p, skb->h.iph = 0x%p, skb->ip_hdr = 0x%p\n", skb->data, skb->h.iph, skb->ip_hdr);
+		
+			printk("ip_forwarding of skb->dev->name %s hopefully equal to %s, %d bytes, to %x\n", skb->dev->name, dev->name, skb->len, target);
+		}
+#endif
+		
+		if (ip_forward(skb, dev, IPFWD_NOTTLDEC, target))
+			kfree_skb(skb, FREE_WRITE);
+		
+		/*
+		 *	Clean up:  We're done with the route and the packet
+		 */
 	 
-	ip_rt_put(rt);
+		/* ip_rt_put(rt); */
  
 
-	/* Record statistics and return */
-	stats->tx_packets++;
-	dev->tbusy=0;
-	return 0;
+		/* Record statistics and return */
+		/* stats->tx_packets++; */
+		dev->tbusy=0;
+		return 0;
+	}
+	else
+	{
+		skb_push(skb, dev->hard_header_len);
+		for (i = 0; i < dev->hard_header_len; i++)
+			skb->data[i] = savedheader[i];
+		ret = prv->hard_start_xmit(skb, dev);
+		dev->tbusy=0;
+
+		return ret;
+	}
+	
 }
 
-static struct enet_statistics *tunnel_get_stats(struct device *dev)
-{
-	return((struct enet_statistics*) dev->priv);
-}
 
 /*
  *	Called when a new tunnel device is initialized.
@@ -495,40 +899,42 @@ ipsec_tunnel_init(struct device *dev)
 	static int tun_msg=0;
 	if(!tun_msg)
 	{
-		printk ( KERN_INFO "ipsec_tunnel: tunnel: version v0.2b2\n" );
+		printk ( KERN_INFO "ipsec_tunnel_init: tunneling code 0.5\n" );
 		tun_msg=1;
 	}
 
 	/* Add our tunnel functions to the device */
 	dev->open		= ipsec_tunnel_open;
 	dev->stop		= ipsec_tunnel_close;
-	dev->hard_start_xmit	= ipsec_tunnel_xmit;
-	dev->get_stats		= tunnel_get_stats;
-	dev->priv = kmalloc(sizeof(struct enet_statistics), GFP_KERNEL);
+	dev->hard_start_xmit	= ipsec_tunnel_start_xmit;
+	dev->get_stats		= ipsec_tunnel_get_stats;
+
+	dev->priv = kmalloc(sizeof(struct ipsecpriv), GFP_KERNEL);
 	if (dev->priv == NULL)
 		return -ENOMEM;
-	memset(dev->priv, 0, sizeof(struct enet_statistics));
+	memset(dev->priv, 0, sizeof(struct ipsecpriv));
 
 	/* Initialize the tunnel device structure */
 	for (i = 0; i < DEV_NUMBUFFS; i++)
 		skb_queue_head_init(&dev->buffs[i]);
 
-	dev->hard_header	= NULL;
-	dev->rebuild_header 	= NULL;
-	dev->set_mac_address 	= NULL;
-	dev->header_cache_bind 	= NULL;
-	dev->header_cache_update= NULL;
+	dev->hard_header	= ipsec_tunnel_hard_header;
+	dev->rebuild_header 	= ipsec_tunnel_rebuild_header;
+	dev->set_multicast_list = NULL;
+	dev->set_mac_address 	= ipsec_tunnel_set_mac_address;
+	dev->do_ioctl		= ipsec_tunnel_ioctl;
+	dev->header_cache_bind 	= ipsec_tunnel_cache_bind;
+	dev->header_cache_update= ipsec_tunnel_cache_update;
 
-	dev->type				= ARPHRD_TUNNEL;
-	dev->hard_header_len 	= TUNL_HLEN;
-	dev->mtu		= 1500-tunnel_hlen; 	/* eth_mtu */
+	dev->type		= ARPHRD_TUNNEL;	/* initially */
+	dev->hard_header_len 	= TUNL_HLEN;		/* plenty */
+	dev->mtu		= 1500-TUNL_HLEN; 	/* eth_mtu */
 	dev->addr_len		= 0;		/* Is this only for ARP? */
-	dev->tx_queue_len	= 2;		/* Small queue */
+	dev->tx_queue_len	= 10;		/* Small queue */
 	memset(dev->broadcast,0xFF, ETH_ALEN);
 
 	/* New-style flags. */
-	dev->flags		= IFF_NOARP;	/* Don't use ARP on this device */
-						/* No broadcasting through a tunnel */
+	dev->flags		= 0;
 	dev->family		= AF_INET;
 	dev->pa_addr		= 0;
 	dev->pa_brdaddr 	= 0;
